package blog;

import org.apache.lucene.analysis.core.WhitespaceAnalyzer;
import org.apache.lucene.queryparser.flexible.core.QueryNodeException;
import org.apache.lucene.util.Version;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * An example to build an index.
 */
public class Example {
    public static void main(String[] av) {

        String[] headers;
        IndexProvider provider = null;
        boolean index = false;

        try {

            File f = new File("index");

            if( ! f.exists() ) {
                if( ! f.mkdir() ) {
                    System.err.println("Could not make directory index");
                    System.exit(1);
                }
                index = true;
            }

            if( ! f.isDirectory() ) {
                System.err.println("Directory index already exists and is not an index");
                System.exit(2);
            }

            provider = new IndexProvider("index", new WhitespaceAnalyzer(Version.LUCENE_40));
            provider.open();
            provider.start();

            //
            // Create the index
            //
            if( index ) {
                System.out.println("Indexing");

                FileReader fr = new FileReader("example.csv");
                BufferedReader br = new BufferedReader(fr);

                ExampleCsvReader csv = new ExampleCsvReader(br);

                headers = csv.readHeader();
                String[] ln;
                Map<String, String> metadata = new HashMap<String, String>();
                while( (ln = csv.readCells()) != null ) {
                    metadata.clear();
                    String content = null;
                    for( int i=0 ; i < headers.length ; i++ ) {
                        if( headers[i].equals(Indexer.CONTENT) ) {
                            if( ln.length > i )
                                content = ln[i];
                            else {
                                System.err.println("Could not index line " + csv.getLineCount());
                            }
                            continue;
                        }
                        metadata.put(headers[i], ln[i]);
                    }

                    ExampleData data = new ExampleData("" + csv.getLineCount(), content, metadata);
                    provider.index(data);
                }

                provider.finish();

            //
            // If not creating the index, run a simple query.
            //
            } else {
                System.out.println("Querying");
                String query = "the";

                provider.createReader();

                runQuery(provider, "the");
                runQuery(provider, "ut eu");
                runQuery(provider, "+ut +f1:two");

            }

        } catch( Throwable t ) {
            System.err.println("Could not process the data");
            t.printStackTrace();
        } finally {
            if( provider != null ) {
                try {
                    provider.close();
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
        }
        try {
            Thread.sleep(1000);
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
    }

    private static void runQuery(IndexProvider provider, String query) throws IOException, QueryNodeException {
        List<String> hits = provider.search(query);

        System.out.println("Query: " + query);
        for( String id : hits ) {
            System.out.println("Hit: " + id);
        }


    }

    public static class ExampleData implements DocumentData {

        private String content;
        private Map<String, String> metadata;
        private String docId;

        public ExampleData(String docId, String content, Map<String, String> metadata) {
            this.content = content;
            this.metadata = metadata;
            this.docId = docId;
        }

        @Override
        public String getContent() {
            return content;
        }

        @Override
        public Map<String, String> getMetadata() {
            return metadata;
        }

        @Override
        public String getDocId() {
            return docId;
        }
    }
}
