package blog;

import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.document.StringField;
import org.apache.lucene.document.TextField;
import org.apache.lucene.index.*;
import org.apache.lucene.queryparser.flexible.core.QueryNodeException;
import org.apache.lucene.queryparser.flexible.standard.StandardQueryParser;
import org.apache.lucene.search.Collector;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.Scorer;
import org.apache.lucene.store.Directory;
import org.apache.lucene.store.NIOFSDirectory;
import org.apache.lucene.util.Version;

import java.io.*;
import java.util.*;

/**
 * A hard implementation allowing for indexing and searching.
 */
public class IndexProvider implements Indexer, Searcher {

    private String dirNm;
    private IndexWriter writer;
    private IndexReader reader;
    private IndexSearcher searcher;
    private Analyzer analyzer;
    private Directory indexDir;

    /**
     * Open the indexing.
     * @param dirNm the directory containing the index, empty if there is an empty index.
     * @param analyzer how the documents are to be indexed and queried.
     */
    public IndexProvider(String dirNm, Analyzer analyzer) {
        this.dirNm = dirNm;
        this.analyzer = analyzer;
    }

    /**
     * Must be called before either indexing or searching can be done.
     * @throws IOException for any system level exceptions
     */
    public void open() throws IOException {
        IndexWriterConfig iwc = new IndexWriterConfig(Version.LUCENE_40, analyzer);
        iwc.setOpenMode(IndexWriterConfig.OpenMode.CREATE_OR_APPEND);
        iwc.setWriteLockTimeout(100);

        indexDir = new NIOFSDirectory(new File(dirNm));
        writer = new IndexWriter(indexDir, iwc);

    }

    /**
     * Must be called to cleanly close off the index.
     * @throws IOException for any system level exceptions
     */
    public void close() throws IOException {
        if( writer != null )
            writer.close();

        if( reader != null )
            reader.close();
    }

    /**
     * Start indexing.
     * @throws IOException
     */
    public void start() throws IOException {
    }

    /**
     * Commit any reads and create a reader.
     * @throws IOException for any underlying exception
     */
    public void finish() throws IOException {
        writer.commit();
        createReader();
    }

    /**
     * If there is already an index of documents, directly create the reader.
     * @throws IOException for any underlying system exceptions
     */
    public void createReader() throws IOException {
        reader = DirectoryReader.open(indexDir);
        searcher = new IndexSearcher(reader);

    }

    /**
     * Index a document's data.
     * @param data the document information
     * @throws IOException
     */
    public void index(DocumentData data) throws IOException {

        Document d = new Document();
        if( data.getDocId() == null || data.getDocId().trim().length() == 0 ) {
            System.err.println("Could not index " + data + " because of null/empty doc id");
            return;
        }


        //
        // This field is stored to allow
        //      index.document(luceneDocId).getField(Indexer.DOC_ID))
        // to return the external DocumentData.getDocId() for the hit list.
        //
        d.add(new TextField(Indexer.DOC_ID, data.getDocId(), Field.Store.YES));

        if( data.getContent() != null ) {
            //
            // For general use this fails because some content will not fit in a single string and must be streamed.
            //
            d.add(new TextField(Indexer.CONTENT, data.getContent(), Field.Store.NO));
        }

        addFields(data, d);

        writer.addDocument(d);

    }


    protected void addFields(DocumentData data, Document d) throws IOException {

        Map<String, String> map = data.getMetadata();

        for( String f : map.keySet() ) {
            //
            // Text field is analyzed.  String field is literal.
            //
            d.add(new TextField(f, map.get(f), Field.Store.NO));
        }
    }


    /**
     * Given a query, return a list of hits.
     * @param query the query string
     * @return a list of document ids
     * @throws IOException for any underlying system exception
     * @throws QueryNodeException for an index exception
     */
    public List<String> search(String query) throws IOException, QueryNodeException {

        if( query == null )
            return new LinkedList<String>();

        if( query.length() == 0 )
            return new LinkedList<String>();

        //
        // This seems a bit slimy to me.  We don't directly use the collector, just use it to house the collection
        // of document ids, which are then used to look up the Indexer.DOC_ID field value.  The collection of those
        // look ups are then returned as the list of hits.
        //
        // This does allow for the underlying index to change and return consistent results.
        // The doc ids are ephemeral.
        //
        List<Integer> docs = new ArrayList<Integer>();
        Collector c = new OutOfOrderCollector(docs);

        StandardQueryParser qp = new StandardQueryParser(analyzer);
        Query q = qp.parse(query, Indexer.CONTENT);

        searcher.search(q, c);

        List<String> ret = new ArrayList<String>(docs.size());
        for( Integer ii : docs ) {
            //
            // Attempt at bullet proofing index hits.  Want to ensure valid hits, as much as possible, at query time.
            //
            if( ii == null )
                continue;

            Document d = reader.document(ii);
            IndexableField f = d.getField(Indexer.DOC_ID);
            if( f == null )
                continue;

            String s = f.stringValue();
            if( s == null )
                continue;

            s = s.trim();
            if( s.length() == 0 )
                continue;

            ret.add(s);
        }

        return ret;
    }

    /**
     * Collect documents, but don't care about sorting them into a ranking order.
     */
    protected static class OutOfOrderCollector extends Collector {

        private List<Integer> docs;
        private int docBase = 0;

        protected OutOfOrderCollector(List<Integer> docs) {
            this.docs = docs;
        }

        @Override
        public void setScorer(Scorer scorer) throws IOException {
        }

        @Override
        public void collect(int doc) throws IOException {
            docs.add(doc + docBase);
        }

        @Override
        public void setNextReader(AtomicReaderContext atomicReaderContext) throws IOException {
            docBase = atomicReaderContext.docBase;
        }

        @Override
        public boolean acceptsDocsOutOfOrder() {
            return true;
        }

    }

}
